`default_nettype none
`timescale 1ns/1ps

module test_mem_read ();

// テスト対象の出力データ型
typedef logic [7:0] data_t;

// テスト対象の信号
logic   clock;      // クロック入力
logic   reset;      // Active Highの同期リセット入力
logic   out_valid;  // 出力VALID
logic   out_ready;  // 出力READY
data_t  out_data;   // 出力データ

// クロック生成 (10ns周期)
initial begin
        clock = 0;
end 
always #(5) begin
    clock = ~clock;
end

// テスト対象のインスタンス化
mem_read dut (.*);

localparam int TEST_COUNTS = 65536; 

/**
 * @summary テスト対象の出力データを受け取る。
 * @param[out] data テスト対象から出力されたデータ
 * @param[in] ready_deassert_factor READYをデアサートする確率。intの範囲 (0~2^32-1) 大きいほどデアサートしやすくなる。
 **/
task automatic get_output_data(output data_t data, input int ready_deassert_factor);
    bit ready_value;
    do begin
        ready_value = $urandom() >= ready_deassert_factor;
        out_ready <= ready_value;
        @(posedge clock);
    end while( !ready_value );
    while(out_valid == 0) @(posedge clock);
    data = out_data;
    out_ready <= ready_value;
endtask

initial begin
    $dumpfile("test-mem_read.vcd"); // VCDファイルを出力する
    $dumpvars(0);                   // 全変数をVCDに出力する
    
    // リセットをアサートして戻す
    reset <= 1;
    out_ready <= 0;
    repeat(2) @(posedge clock);
    reset <= 0;
    @(posedge clock);
    while(!out_valid) @(posedge clock); // 最初のout_validアサートまで待つ

    fork
        begin
            int phase_counter;
            phase_counter = 0;
            // 65536回分の出力を確認する
            for(int i = 0; i < TEST_COUNTS; i++) begin
                for(int phase = 0; phase < 2; phase++) begin
                    data_t expected_data;
                    data_t actual_data;

                    expected_data  = (phase == 0) ? i[7:0] : i[15:8];    // phase 0のときは下位バイト、phase 1のときは上位バイト
                    get_output_data(actual_data, 32'h80000000); // データを読み出す

                    // 期待値チェック
                    if( actual_data !== expected_data ) $error("#%04d data mismatch. expected: %02x, actual: %02x", i, expected_data, out_data);
                end
            end
            // ここでは最後のVALIDアサートサイクルのはずなので1サイクル進める。
            @(posedge clock);
            // 直後10サイクルはout_validがアサートされないことを確認
            for(int i = 0; i < 10; i++) begin
                // 期待値チェック
                if( out_valid ) $error("out_valid must be deasserted after output ends.");
                @(posedge clock);
            end
        end
        begin
            // タイムアウト検出。余裕を見て 65536 * 3 * 2 + 20 サイクル以内でテストが完了することを期待する。
            static int limit = TEST_COUNTS * 3 * 2 + 20;
            for(int cycles = 0; cycles < limit; cycles++ ) @(posedge clock);
            $error("Error: simulation timed out after %d cycles.", limit);
        end
    join_any
    disable fork;
    $finish;
end

endmodule

`default_nettype wire